import time
from enum import Enum
from typing import Union, Dict, List, Any
from pydantic import BaseModel


class RunMode(Enum):
    # 批采
    BATCH = 'batch'
    # 流采
    STREAM = 'stream'


class RabbitMeta(BaseModel):
    """
    消息发布元数据
    """
    # 交换机键
    exchange_key: str = None
    # 交换机类型
    exchange_type: str = None
    # 绑定关系（路由）键
    routing_key: str = None
    # 队列名称
    queue_key: str = None

    def __init__(self, exchange_key: str, exchange_type: str, routing_key: str, queue_key: str = None, **data: Any):
        super().__init__(**data)
        self.exchange_key = exchange_key
        self.exchange_type = exchange_type
        self.routing_key = routing_key
        self.queue_key = queue_key


class MsgType(Enum):
    """
    消息类型枚举
    """
    # 爬虫控制指令
    ACT_OPENED = 'OPENED'
    ACT_CLOSED = 'CLOSED'
    # 数据消息
    SKU_TEXT = 'SKU_TEXT'
    SKU_IMAGES = 'SKU_IMAGES'


class SignalType(Enum):
    """
    采云间爬虫状态信号枚举
    """
    OPENED = 'OPENED'
    CLOSED = 'CLOSED'


class BaseMsg(BaseModel):
    """
    标准消息数据模型
    """
    # 消息接收者应用编码
    app_code: str = None
    # 消息类型（控制消息：ACT；数据消息：DATA；流式数据消息：STREAM）
    msg_type: Union[str, MsgType]
    # 消息体
    msg_body: Union[str, Dict, List] = None
    # 产生时间戳
    gen_time: int = int(time.time())



# -*- coding: utf-8 -*-
from typing import List
from ..util import logger
from ..dao import spider as spider_dao

LOGGER = logger.get('批次爬虫')


def get_spider(spider_id: str):
    row = spider_dao.get_spider(spider_id=spider_id)
    return row


def get_spiders_by_id(spider_ids: List[str]):
    spiders = spider_dao.get_spiders_by_id(spider_ids)
    return spiders



# -*- coding: utf-8 -*-
import json
import random
import uuid
from typing import List, Dict
from zcbot_url_parser import parser as url_parser
from ..model.enums import CommonStatus, TaskMode
from ..util import logger, cfg
from ..util import time as time_lib
from ..util.exceptions import BizException
from ..client.redis_client import Redis
from ..util import redis_key as redis_key_util
from ..model.stream import StreamApiData, StreamTaskItem
from ..dao import spider as spider_dao
from ..dao.meta import create_pipeline, run_pipeline

LOGGER = logger.get('流采')


def publish(api_data: StreamApiData, attach_spiders_dict: Dict):
    """
    准备批次采集任务物料
    """
    # 如果上游传递则直接使用，否则本系统生成批次编号
    app_code = api_data.appCode
    task_items = api_data.taskItems or []
    spider_ids = api_data.spiderIds or []
    # 默认通用通道
    channel_code = api_data.channelCode or 'common'
    if not app_code:
        raise BizException(f'[分拣]app_code未指定 api_data={api_data}')
    if not task_items:
        raise BizException(f'[分拣]任务清单为空 api_data={api_data}')
    if not spider_ids:
        raise BizException(f'[分拣]未指定spider_ids api_data={api_data}')
    spider_ids_string = ",".join(spider_ids)
    LOGGER.info(f'[分拣]开始流采分拣 app_code={app_code}, spider_ids={spider_ids_string}, rows={len(task_items)}')

    # 物料分拣
    result_map, classify_result_list, classify_ignore_list = classify(api_data)
    # 初始化采集任务
    init_job(result_map, spider_ids, channel_code, attach_spiders_dict)
    LOGGER.info(f'[分拣]流采分拣完成 app_code={app_code}, spider_ids={spider_ids_string}, rows={len(task_items)}')

    return result_map, classify_result_list, classify_ignore_list


def classify(api_data: StreamApiData) -> (Dict[str, List[StreamTaskItem]], List[dict]):
    """
    任务分拣
    :param api_data: 待分拣任务清单，
    :return:
    """
    app_code = api_data.appCode
    todo_list = api_data.taskItems
    LOGGER.info('[分拣]开始流采分拣 total=%s' % len(todo_list))

    result_map = dict()
    classify_result_list = list()
    classify_ignore_list = list()
    for idx, row in enumerate(todo_list):
        plat_code = row.platCode
        # 如果有plat_code则不分拣(兼容购物党等平台)
        if not plat_code:
            # 解析链接组装任务
            url_model = url_parser.parse_url(row.url)
            if not url_model:
                LOGGER.warning(f'[分拣]未识别的流采任务 sn={row.sn} url={row.url}')
                classify_ignore_list.append(row)
                continue
            link_id, plat_code, plat_name, ec_sku_id = url_model.link_sn, url_model.plat_code, url_model.plat_name, url_model.ec_sku_id
            if link_id and plat_code and ec_sku_id:
                row.linkId = link_id
                row.platCode = plat_code
                row.platName = plat_name
                row.ecSkuId = ec_sku_id
            else:
                LOGGER.warning(f'[分拣]未识别的流采任务 url={row.url}, url_sn={link_id}, plat_code={plat_code}, ec_sku_id={ec_sku_id}')
                classify_ignore_list.append(row)
                continue
        # 额外字段
        row.appCode = row.appCode or app_code
        row.rowId = row.rowId or row.sn
        # 分拣
        if row.sn and plat_code:
            task_list = result_map.get(plat_code)
            if not task_list:
                task_list = list()
                result_map[plat_code] = task_list
            task_list.append(row)
            classify_result_list.append(row.to_classify_result())
    # 统计
    count_map = dict()
    for key in result_map.keys():
        count_map[key] = len(result_map.get(key))
    LOGGER.info('[分拣]流采分拣完成 %s' % count_map)

    return result_map, classify_result_list, classify_ignore_list


def init_job(result_map: Dict[str, List], spider_ids: List[str], channel_code: str, attach_spiders_dict: Dict) -> (Dict[str, List[StreamTaskItem]], List[dict], str):
    for spider_id in spider_ids:
        # 根据spider查询plat_code
        spider_conf = spider_dao.get_spider(spider_id, CommonStatus.ON)
        if not spider_conf:
            LOGGER.warning(BizException(f'[初始化队列]{spider_id}爬虫配置表zcbot_spider异常或不存在'))
            continue
        else:
            # 初始化队列
            plat_code = spider_conf.platCode
            task_items = result_map.get(plat_code)
            if task_items:
                for task_item in task_items:
                    callback = task_item.get("callback", {})
                    if callback:
                        callback.update("attachSpiders", attach_spiders_dict.get(spider_id))
                    else:
                        callback = {"attachSpiders": attach_spiders_dict.get(spider_id)}
                    task_item['callback'] = callback
                LOGGER.info(f'[初始化队列]平台编码:{plat_code}开始初始化数据')
                redis_key = redis_key_util.get_stream_task_queue_key(spider_id=spider_id, channel_code=channel_code)
                plat_init_count = init_task_queue(redis_key, spider_conf, task_items)
                LOGGER.info(f'[发布]初始化流采队列 plat_code={plat_code}, plat_init_count={plat_init_count}')
            else:
                pass
                # LOGGER.warning(f'[初始化队列]平台编码:{plat_code}无初始化数据')


def init_task_queue(redis_key, spider_conf, rows):
    """
    初始化任务队列
    :param redis_key:
    :param spider_conf:
    :param rows:
    :return:
    """
    # 初始化队列
    task_mode = spider_conf.taskMode
    LOGGER.info(f'[队列]任务队列初始化开始 redis_key={redis_key}, task_mode={task_mode}')
    if task_mode and redis_key and task_mode == TaskMode.batch.value:
        # 批量模式（如：京东价格）
        batch_size = spider_conf.batchSize
        return _init_redis_batch(redis_key, rows, batch_size)
    else:
        # 单条采集模式
        return _init_redis(redis_key, rows)


def add_to_task_mapper(req_id, task_data, task_queue_key, pipe):
    """
    加入任务映射队列，用于重试任务源数据
    """
    _data = {
        'queue': task_queue_key,
        'source': task_data,
        'genTime': time_lib.current_timestamp10()
    }
    _expire_seconds = cfg.get_int('ZCBOT_CORE_REDIS_QUEUE_EXPIRE') or 12 * 3600
    pipe.set(redis_key_util.get_retry_request_source_key(req_id), json.dumps(_data), ex=_expire_seconds)


def _init_redis(redis_key, rows: List[StreamTaskItem]):
    rds = Redis()
    pipe = create_pipeline()
    for row in rows:
        req_id = str(uuid.uuid4())
        task = {
            'reqId': req_id,
            "sn": str(row.sn),
            "url": row.url,
            "platCode": row.platCode,
            "rowId": str(row.rowId),
            "appCode": row.appCode,
        }
        if row.callback:
            task["callback"] = row.callback
        if row.itemConfig:
            task["itemConfig"] = row.itemConfig

        # 任务入队
        pipe.lpush(redis_key, json.dumps(task))
        # 加入重试源数据集合
        add_to_task_mapper(req_id, task, redis_key, pipe)
    run_pipeline(pipe)
    count = rds.client.llen(redis_key)
    LOGGER.info(f'[队列]任务队列初始化完成 -> 单条模式 key={redis_key}, row={count}, count={len(rows)}')

    return count


def _init_redis_batch(redis_key, rows: List[StreamTaskItem], batch_size):
    _expire_seconds = cfg.get_int('ZCBOT_CORE_REDIS_QUEUE_EXPIRE') or 12 * 3600
    rds = Redis()
    batch_list = []
    random.shuffle(rows)
    pipe = create_pipeline()
    for row in rows:
        task = {
            "sn": str(row.sn),
            "url": row.url,
            "platCode": row.platCode,
            "rowId": str(row.rowId),
            "appCode": row.appCode,
            "callback": row.callback,
            "itemConfig": row.itemConfig
        }

        batch_list.append(task)
        if len(batch_list) >= batch_size:
            req_id = str(uuid.uuid4())
            batch_row_data = {
                'reqId': req_id,
                'data': batch_list
            }
            # 任务入队
            pipe.lpush(redis_key, json.dumps(batch_row_data))
            # 加入重试源数据集合
            add_to_task_mapper(req_id, batch_row_data, redis_key, pipe)
            batch_list = []
    if batch_list:
        req_id = str(uuid.uuid4())
        batch_row_data = {
            'reqId': req_id,
            'data': batch_list
        }
        # 任务入队
        pipe.lpush(redis_key, json.dumps(batch_row_data))
        # 加入重试源数据集合
        add_to_task_mapper(req_id, batch_row_data, redis_key, pipe)
    run_pipeline(pipe)
    count = rds.client.llen(redis_key)
    LOGGER.info(f'[队列]任务队列初始化完成 -> 批量模式 key={redis_key}, row={count}, count={len(rows)}')

    return len(rows)



from pika.exchange_type import ExchangeType
from .model import RabbitMeta, RunMode


# rabbit配置信息统一定义（同spider中配置一致）
def get_rabbit_result_keys(run_mode: RunMode, app_code: str) -> RabbitMeta:
    if run_mode == RunMode.BATCH:
        # 批采
        exchange_key = f'zcbot.batch_result'
        exchange_type = ExchangeType.topic
        routing_key = f'zcbot.batch_result.{app_code}'
        queue_key = f'zcbot.batch_result.{app_code}'
        return RabbitMeta(exchange_key, exchange_type, routing_key, queue_key)
    elif run_mode == RunMode.STREAM:
        # 流采
        exchange_key = f'zcbot.stream_result'
        exchange_type = ExchangeType.topic
        routing_key = f'zcbot.stream_result.{app_code}'
        queue_key = f'zcbot.stream_result.{app_code}'
        return RabbitMeta(exchange_key, exchange_type, routing_key, queue_key)
    else:
        # 死信队列
        exchange_key = f'zcbot.dead_result'
        exchange_type = ExchangeType.topic
        routing_key = f'zcbot.dead_result.dead_routing'
        queue_key = f'zcbot.dead_result.dead_queue'
        return RabbitMeta(exchange_key, exchange_type, routing_key, queue_key)



# -*- coding: UTF-8 -*-


# 【批采】采集任务队列
def get_batch_task_queue_key(batch_id: str, spider_id: str):
    return f'zcbot:batch:{batch_id}:{spider_id}'


# 【流采】采集任务队列
def get_stream_task_queue_key(spider_id: str, channel_code: str):
    return f'zcbot:stream:{spider_id}:{channel_code}'


# # 采集任务队列
# def get_task_queue_key_for_remove(batch_id: str):
#     return f'zcbot:{batch_id}:*'


# 【重试】采集重试任务源数据映射
def get_retry_request_source_key(req_id: str):
    return f'zcbot:retry:source:{req_id}'


# 测试
if __name__ == '__main__':
    pass



# -*- coding: utf-8 -*-
import json
import logging
import threading
import time
import pika
import traceback
from pika.exceptions import StreamLostError
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, Union, List
from .key_builder import get_rabbit_result_keys
from .model import RabbitMeta, RunMode
from .processor import AbstractMessageProcess
from ..util.decator import singleton

LOGGER = logging.getLogger(__name__)


class Rabbit(object):
    """
    通用Rabbit客户端
    """
    lock = threading.Lock()

    def __init__(self, connection_uri=None):
        self.connection_uri = connection_uri
        self.connection = pika.BlockingConnection(pika.URLParameters(self.connection_uri))
        self.default_channel = self.connection.channel()
        LOGGER.info(f'Rabbit初始化链路 connection_uri -> {self.connection_uri}')

    def ensure_channel(self, force: bool = False):
        """
        链路检查，确保链路通畅
        :return:
        """
        # 双重校验
        if force or not self.connection or self.connection.is_closed or not self.default_channel or self.default_channel.is_closed:
            try:
                self.lock.acquire(timeout=30)
                if force or not self.connection or self.connection.is_closed or not self.default_channel or self.default_channel.is_closed:
                    # 清理链路
                    if self.default_channel:
                        try:
                            self.default_channel.close()
                        except Exception as e:
                            LOGGER.error(e)
                    if self.connection:
                        try:
                            self.connection.close()
                        except Exception as e:
                            LOGGER.error(e)
                    # 重新建立链接
                    self.connection = pika.BlockingConnection(pika.URLParameters(self.connection_uri))
                    self.default_channel = self.connection.channel()
                    LOGGER.info(f'重建链路... -> {self.connection_uri}')
            finally:
                if self.lock.locked():
                    self.lock.release()

    def close(self):
        if self.default_channel and self.default_channel.is_open:
            try:
                self.default_channel.close()
            except Exception as e:
                LOGGER.error(e)
        if self.connection and self.connection.is_open:
            try:
                self.connection.close()
            except Exception as e:
                LOGGER.error(e)

    # 销毁关闭链接
    def __del__(self):
        try:
            self.close()
        except Exception as e:
            LOGGER.error(e)


@singleton
class RabbitResultReceiver(Rabbit):
    """
    【单例】异步采集结果接收
    1、通道随批次创建，随采集完成自动销毁
    2、消息与队列自动创建自动删除
    3、通道按批次区分
    """
    declare_lock = threading.Lock()

    def __init__(self, processor: Union[AbstractMessageProcess, Callable], connection_uri: str, max_watcher_count: int = None, max_processor_count: int = None):
        super().__init__(connection_uri)
        # 默认值，和zcbot_spider保持一致
        _max_watcher_count = max_watcher_count or 2
        _max_processor_count = max_processor_count or 16

        self.processor = processor
        self.watcher_executor = ThreadPoolExecutor(max_workers=_max_watcher_count, thread_name_prefix='rabbit-watcher')
        self.processor_executor = ThreadPoolExecutor(max_workers=_max_processor_count, thread_name_prefix='rabbit-processor')
        # 路由映射
        self.queue_declared_filter = set()
        LOGGER.info(f'[队列]监听器初始化 connection_uri={self.connection_uri}')

    def watch(self, run_mode: RunMode, app_code: Union[str, List[str]]):
        # rabbit配置信息统一构建【批采+流采】
        if isinstance(app_code, str):
            # 单个应用监测
            rabbit_meta = get_rabbit_result_keys(run_mode, app_code=app_code)
            self.watcher_executor.submit(self._receive_message, rabbit_meta)
        elif isinstance(app_code, List):
            # 多应用监测
            for _app_code in app_code:
                rabbit_meta = get_rabbit_result_keys(run_mode, app_code=_app_code)
                self.watcher_executor.submit(self._receive_message, rabbit_meta)

    def ensure_declares(self, meta: RabbitMeta):
        # 链路检查
        self.ensure_channel()
        # 双重校验
        if meta.queue_key not in self.queue_declared_filter:
            try:
                self.declare_lock.acquire(timeout=15)
                if meta.queue_key not in self.queue_declared_filter:
                    # 初始化消息队列
                    # 定义交换机
                    self.default_channel.exchange_declare(exchange=meta.exchange_key, exchange_type=meta.exchange_type, durable=True)
                    # 定义消息队列
                    self.default_channel.queue_declare(queue=meta.queue_key, durable=True)
                    # 定义绑定关系（路由）
                    self.default_channel.queue_bind(queue=meta.queue_key, exchange=meta.exchange_key, routing_key=meta.routing_key)
                    # 避免重复初始化
                    self.queue_declared_filter.add(meta.queue_key)
                    LOGGER.info('===============================')
                    LOGGER.info(f'init exchange_key -> {meta.exchange_key}')
                    LOGGER.info(f'init rabbit_routing_key -> {meta.routing_key}')
                    LOGGER.info(f'init rabbit_queue -> {meta.queue_key}')
                    LOGGER.info('===============================')
            finally:
                if self.declare_lock.locked():
                    self.declare_lock.release()

    def _receive_message(self, rabbit_meta: RabbitMeta):
        """
        连接消息队列并启动消费，阻塞队列（需要独立线程运行或挂在后台任务运行）
        """
        LOGGER.info(f'[队列]开始接收')
        _process_func = self.processor
        if isinstance(self.processor, AbstractMessageProcess):
            _process_func = self.processor.process_message
        try:
            self.ensure_declares(rabbit_meta)
            # 接收
            for method, properties, body in self.default_channel.consume(rabbit_meta.queue_key, auto_ack=False):
                # 保持通道激活状态
                if not method or not properties or not properties.headers:
                    LOGGER.error(f'[队列]无效消息 method={method}, properties={properties}, body={body}')
                    continue
                # 消息解析并发处理
                msg_type = properties.headers.get('msg_type', None)
                msg_tag = properties.headers.get('msg_tag', None)
                body_json = json.loads(body.decode())
                self.processor_executor.submit(_process_func, msg_type, body_json, msg_tag)
                # 消息确认
                self.default_channel.basic_ack(method.delivery_tag)
        except (StreamLostError, ConnectionAbortedError):
            LOGGER.error(f'[队列]服务端关闭链接通道StreamLostError,ConnectionAbortedError -> 重连 {traceback.format_exc()}')
        except pika.exceptions.ConnectionClosedByBroker:
            LOGGER.error(f'[队列]链接关闭异常ConnectionClosedByBroker -> 重连 {traceback.format_exc()}')
        except pika.exceptions.AMQPChannelError:
            LOGGER.error(f'[队列]通道关闭异常AMQPChannelError -> 重连 {traceback.format_exc()}')
        except pika.exceptions.AMQPConnectionError:
            LOGGER.error(f'[队列]链接关闭异常AMQPConnectionError -> 重连 {traceback.format_exc()}')
        except Exception:
            LOGGER.error(f'[队列]接收过程异常 -> 重连 {traceback.format_exc()}')
        finally:
            try:
                # 清空队列定义过滤校验，防止断链后三要素丢失
                self.queue_declared_filter.clear()
                # 关闭链接和通道（链接关闭通道自动关闭）
                self.default_channel.close()
                self.connection.close()
                LOGGER.info(f'[队列]销毁队列')
            except Exception:
                LOGGER.error(f'[队列]关闭链接异常 {traceback.format_exc()}')

            # 递归重试
            time.sleep(10)
            LOGGER.warning(f'[队列]异常重试中...')
            self._receive_message(rabbit_meta)



# -*- coding: utf-8 -*-
import logging
import threading
import traceback
from typing import Union, Dict, List
from .model import MsgType

LOGGER = logging.getLogger(__name__)


class AbstractMessageProcess(object):
    """
    抽象消息处理器
    """

    # 消息处理入口
    def process_message(self, msg_type: str, msg_data: Union[Dict, List] = None, msg_tag: str = None):
        try:
            if msg_type == MsgType.SKU_TEXT.name:
                self.process_sku_text(msg_data=msg_data, msg_tag=msg_tag)
            elif msg_type == MsgType.SKU_IMAGES.name:
                self.process_sku_images(msg_data=msg_data, msg_tag=msg_tag)
            elif msg_type == MsgType.ACT_OPENED.name:
                self.process_opened_action(msg_data=msg_data, msg_tag=msg_tag)
            elif msg_type == MsgType.ACT_CLOSED.name:
                self.process_closed_action(msg_data=msg_data, msg_tag=msg_tag)
            else:
                self.process_others(msg_data=msg_data, msg_type=msg_type, msg_tag=msg_tag)
        except Exception:
            LOGGER.error(f'[流采消息]解析异常 {traceback.format_exc()}')
            LOGGER.error(f'[流采消息]解析异常 msg_data={msg_data}, except={traceback.format_exc()}')

    def process_sku_text(self, msg_data: Union[Dict, List], msg_tag: str):
        LOGGER.warning(f'[{threading.current_thread().name}]process: msg_type=process_sku_text, msg_data={msg_data}, msg_tag={msg_tag}')

    def process_sku_images(self, msg_data: Union[Dict, List], msg_tag: str):
        LOGGER.warning(f'[{threading.current_thread().name}]process: msg_type=process_sku_images, msg_data={msg_data}, msg_tag={msg_tag}')

    def process_opened_action(self, msg_data: Union[Dict, List], msg_tag: str):
        LOGGER.warning(f'[{threading.current_thread().name}]process: msg_type=process_opened_action, msg_data={msg_data}, msg_tag={msg_tag}')

    def process_closed_action(self, msg_data: Union[Dict, List], msg_tag: str):
        LOGGER.warning(f'[{threading.current_thread().name}]process: msg_type=process_closed_action, msg_data={msg_data}, msg_tag={msg_tag}')

    def process_others(self, msg_data: Union[Dict, List], msg_type: str, msg_tag: str):
        LOGGER.error(f'[流采消息]未知类型消息 msg_type={msg_type}, msg_data={msg_data}, msg_tag={msg_tag}')
